@@ -7,20 +7,8 @@ __license__ = "MIT"
 import importlib
 import logging
 import posixpath
-
-# Python 3 compatibility:
-try:
-    import httplib
-except ImportError:  # py3
-    from http import client as httplib
-try:
-    import urlparse
-except ImportError:  # py3
-    from urllib import parse as urlparse
-try:
-    basestring
-except NameError:  # py3
-    basestring = (bytes, str)
+from http import client as httplib
+from urllib import parse as urlparse
 
 logger = logging.getLogger(__name__)
 logger.setLevel(level=logging.WARNING)
@@ -41,7 +29,7 @@ SERVICES = {
         "supported_https_auths": ["password", "token"],
         "default_port": 8006,
         "token_separator": "=",
-        "ssh_additional_options": ["--output-format", "json"],
+        "cli_additional_options": ["--output-format", "json"],
     },
     "PMG": {
         "supported_backends": ["local", "https", "openssh", "ssh_paramiko"],
@@ -61,7 +49,47 @@ def config_failure(message, *args):
     raise NotImplementedError(message.format(*args))
 
 
-class ProxmoxResourceBase(object):
+class ResourceException(Exception):
+    """
+    An Exception thrown when an Proxmox API call failed
+    """
+
+    def __init__(self, status_code, status_message, content, errors=None):
+        """
+        Create a new ResourceException
+
+        :param status_code: The HTTP status code (faked by non-HTTP backends)
+        :type status_code: int
+        :param status_message: HTTP Status code (faked by non-HTTP backends)
+        :type status_message: str
+        :param content: Extended information on what went wrong
+        :type content: str
+        :param errors: Any specific errors that were encountered (converted to string), defaults to None
+        :type errors: Optional[object], optional
+        """
+        self.status_code = status_code
+        self.status_message = status_message
+        self.content = content
+        self.errors = errors
+        if errors is not None:
+            content += f" - {errors}"
+        message = f"{status_code} {status_message}: {content}".strip()
+        super().__init__(message)
+
+
+class AuthenticationError(Exception):
+    def __init__(self, msg):
+        super().__init__(msg)
+        self.msg = msg
+
+    def __str__(self):
+        return self.msg
+
+
+class ProxmoxResource:
+    def __init__(self, **kwargs):
+        self._store = kwargs
+
     def __getattr__(self, item):
         if item.startswith("_"):
             raise AttributeError(item)
@@ -74,31 +102,14 @@ class ProxmoxResourceBase(object):
     def url_join(self, base, *args):
         scheme, netloc, path, query, fragment = urlparse.urlsplit(base)
         path = path if len(path) else "/"
-        path = posixpath.join(path, *[("%s" % x) for x in args])
+        path = posixpath.join(path, *[str(x) for x in args])
         return urlparse.urlunsplit([scheme, netloc, path, query, fragment])
 
-
-class ResourceException(Exception):
-    def __init__(self, status_code, status_message, content, errors=None):
-        self.status_code = status_code
-        self.status_message = status_message
-        self.content = content
-        self.errors = errors
-        if errors is not None:
-            content += " - {0}".format(errors)
-        message = "{0} {1}: {2}".format(status_code, status_message, content).strip()
-        super(ResourceException, self).__init__(message)
-
-
-class ProxmoxResource(ProxmoxResourceBase):
-    def __init__(self, **kwargs):
-        self._store = kwargs
-
     def __call__(self, resource_id=None):
-        if not resource_id:
+        if resource_id in (None, ""):
             return self
 
-        if isinstance(resource_id, basestring):
+        if isinstance(resource_id, (bytes, str)):
             resource_id = resource_id.split("/")
         elif not isinstance(resource_id, (tuple, list)):
             resource_id = [str(resource_id)]
@@ -112,11 +123,26 @@ class ProxmoxResource(ProxmoxResourceBase):
     def _request(self, method, data=None, params=None):
         url = self._store["base_url"]
         if data:
-            logger.info("%s %s %r", method, url, data)
+            logger.info(f"{method} {url} {data}")
         else:
-            logger.info("%s %s", method, url)
-        resp = self._store["session"].request(method, url, data=data or None, params=params)
-        logger.debug("Status code: %s, output: %s", resp.status_code, resp.content)
+            logger.info(f"{method} {url}")
+
+        # passing None values to pvesh command breaks it, let's remove them just as requests library does
+        # helpful when dealing with function default values higher in the chain, no need to clean up in multiple places
+        if params:
+            # remove keys that are set to None
+            params_none_keys = [k for (k, v) in params.items() if v is None]
+            for key in params_none_keys:
+                del params[key]
+
+        if data:
+            # remove keys that are set to None
+            data_none_keys = [k for (k, v) in data.items() if v is None]
+            for key in data_none_keys:
+                del data[key]
+
+        resp = self._store["session"].request(method, url, data=data, params=params)
+        logger.debug(f"Status code: {resp.status_code}, output: {resp.content}")
 
         if resp.status_code >= 400:
             if hasattr(resp, "reason"):
@@ -126,7 +152,7 @@ class ProxmoxResource(ProxmoxResourceBase):
                         resp.status_code, ANYEVENT_HTTP_STATUS_CODES.get(resp.status_code)
                     ),
                     resp.reason,
-                    errors=(self._store["serializer"].loads_errors(resp) or {}),
+                    errors=(self._store["serializer"].loads_errors(resp)),
                 )
             else:
                 raise ResourceException(
@@ -160,7 +186,7 @@ class ProxmoxResource(ProxmoxResourceBase):
 
 class ProxmoxAPI(ProxmoxResource):
     def __init__(self, host=None, backend="https", service="PVE", **kwargs):
-        super(ProxmoxAPI, self).__init__(**kwargs)
+        super().__init__(**kwargs)
         service = service.upper()
         backend = backend.lower()
 
@@ -170,7 +196,7 @@ class ProxmoxAPI(ProxmoxResource):
 
         # throw error for unsupported backend for service
         if backend not in SERVICES[service]["supported_backends"]:
-            config_failure("{} backend does not support {} backend", service, backend)
+            config_failure("{} service does not support {} backend", service, backend)
 
         if host is not None:
             if backend == "local":
@@ -181,7 +207,7 @@ class ProxmoxAPI(ProxmoxResource):
         kwargs["service"] = service
 
         # load backend module
-        self._backend = importlib.import_module(".backends.%s" % backend, "proxmoxer").Backend(
+        self._backend = importlib.import_module(f".backends.{backend}", "proxmoxer").Backend(
             **kwargs
         )
         self._backend_name = backend
